# Dec 15, 2018
# Koji Ota

package require Tk
package require BWidget
package require ctext
package require TclOO
package require Img

load imgop.dll

#option add *Font {{Meiryo UI} 9}

proc config { name args } {
	global _config_
	if {[llength $args] > 0} {
		set _config_($name) [lindex $args 0]
	}
	return $_config_($name)
}

config dot "c:/Program Files (x86)/Graphviz2.38/bin/dot.exe"
config path.projects "./projects"

set uuidbase [clock clicks -milliseconds]
set uuidcount 1
proc generate_uuid { } {
	set uuid "UID:${::uuidbase}::${::uuidcount}"
	incr ::uuidcount
	return $uuid
}
oo::class create Data {
	variable m_uuid
	variable m_data
	
	constructor args {
		set m_uuid [generate_uuid]
	}
	
	method uuid { } {
		return $m_uuid
	}
	
	method set { name value } {
		set m_data($name) $value
	}
	
	method get { name } {
		return $m_data($name)
	}
}

oo::class create ImageViewer {
	variable m_mapped
	variable m_zoom
	variable m_imgfile
	variable m_client_w
	variable m_client_h
	variable m_px
	variable m_py
	variable m_xpan
	variable m_ypan
	
	constructor args {
		set m_mapped 0
		set m_zoom 1.0
		set m_imgfile {}
		set m_client_w 0
		set m_client_h 0
		set m_px 0
		set m_py 0
		set m_xpan 0
		set m_ypan 0
	}
	
	method create { } {
		toplevel .iv
		wm protocol .iv WM_DELETE_WINDOW "wm withdraw .iv"
		wm geometry .iv "1024x768"
		
		pack [frame .iv.mb] -fill x
		
		set m .iv.mb.file.m
		pack [menubutton .iv.mb.file -text "File" -menu $m] -side left
		menu $m -tearoff 0
		$m add command -label "Load" -command "[self] selectFile"
		set m .iv.mb.view.m
		pack [menubutton .iv.mb.view -text "View" -menu $m] -side left
		menu $m -tearoff 0
		$m add command -label "Reset" -command "[self] reset"
		
		image create photo imageFore
		pack [label .iv.view -image imageFore] -fill both -expand 1
	
		bind .iv <MouseWheel> "[self] wheel %D"
		bind .iv <Configure> "[self] resize"
		bind .iv <ButtonPress-1> "[self] press %x %y"
		bind .iv <B1-Motion> "[self] motion %x %y"
		bind .iv <Control-r> "[self] reset"
		bind .iv <Control-a> "[self] wheel -120"
		bind .iv <Control-z> "[self] wheel 120"
		bind .iv <Control-Up> "[self] pan 0 20"
		bind .iv <Control-Down> "[self] pan 0 -20"
		bind .iv <Control-Left> "[self] pan -20 0"
		bind .iv <Control-Right> "[self] pan 20 0"
	
		set m_mapped 1
	}
	
	method reset { } {
		set m_zoom 1.0
		set m_xpan 0.0
		set m_ypan 0.0
		[self] draw
	}
	
	method resize { } {
		set client_w [winfo width .iv.view]
		set client_h [winfo height .iv.view]
	
		if {$client_w == $m_client_w && $client_h == $m_client_h} {
			return
		}
		set m_client_w $client_w
		set m_client_h $client_h
		imageFore blank
		imageFore configure -width $m_client_w -height $m_client_h
		
		[self] adjust_zoom
		[self] draw
	}
	
	method selectFile { } {
		set file [tk_getOpenFile -initialdir [pwd]]
		if {$file == {}} {
			return
		}
		
		[self] show $file
	}
	
	method show { {file {}} } {
		if {$m_mapped == 0} {
			[self] create
		}
		wm deiconify .iv
		
		if {$file != {}} {
			[self] load $file
		}
	}
	
	method load { file } {
		set m_imgfile $file
		
		set names [image names]
		if {[lsearch $names imageBack] >= 0} {
			image delete imageBack
		}
		
		image create photo imageBack -file $m_imgfile
		set m_zoom 1.0
		
		[self] adjust_zoom
		[self] draw
	}
	
	method adjust_zoom { } {
		if {$m_client_w == 0 && $m_client_h == 0} {
			return
		}
		if {$m_imgfile == {}} {
			return
		}
	
		set iw [image width imageBack]
		set ih [image height imageBack]
		
		if {$iw > $ih} {
			# base is iw
			set m_zoom [expr ($m_client_w+0.0)/($iw+0.0)]
		} else {
			# base is ih
			set m_zoom [expr ($m_client_h+0.0)/($ih+0.0)]
		}
	}
	
	method draw { } {
		if {$m_imgfile == {}} {
			return
		}
		
		imgop::transform imageBack imageFore -scale $m_zoom -interpolation linear -xpan $m_xpan -ypan $m_ypan
	}
	
	method wheel { d } {
		if {$d > 0} {
			set m_zoom [expr $m_zoom * 0.8]
		} else {
			set m_zoom [expr $m_zoom * 1.2]
		}
		[self] draw
	}
	
	method press { x y } {
		set m_px $x
		set m_py $y
	}
	
	method motion { x y } {
		incr m_xpan [expr $x - $m_px]
		incr m_ypan [expr $y - $m_py]
		set m_px $x
		set m_py $y
		[self] draw
	}
	
	method pan { dx dy } {
		incr m_xpan $dx
		incr m_ypan $dy
		[self] draw
	}
}

oo::class create Project {
	variable m_app
	variable m_tree
	variable m_editor
	variable m_mapped
	variable m_active
	variable m_input
	
	constructor {app w} {
		set m_app $app
		set m_tree $w
		set m_mapped 0
		
		$m_tree bindText <Double-1> "[self] open %X %Y"
		$m_tree bindImage <Double-1> "[self] open %X %Y"
		$m_tree bindText <3> "[self] popup %X %Y"
		$m_tree bindImage <3> "[self] popup %X %Y"
		
		set m .menu_project
		menu $m -tearoff 0
		$m add command -label "New File" -command "[self] newFileDialog"
	}
	
	method showDialog { } {
		if {$m_mapped == 0} {
			toplevel .project
			wm protocol .project WM_DELETE_WINDOW "wm withdraw .project"
			pack [frame .project.f1] -fill x -padx 10 -pady 10
			pack [label .project.f1.w1 -text "New Project Name:"] -side left
			pack [entry .project.f1.w2 -textvariable [my varname m_input]] -side left
			pack [frame .project.f2] -fill x
			pack [button .project.f2.w2 -width 8 -text "Cancel" -command "wm withdraw .project"] -side right
			pack [button .project.f2.w1 -width 8 -text "OK" -command "[self] newProject"] -side right
		}
		set m_input ""
		wm deiconify .project
	}
	
	method newProject { } {
		if [file exists $m_input] {
			return
		}
		
		set dir "[config path.projects]/$m_input"
		file mkdir $dir
		
		set data [Data new]
		$data set type folder
		$data set path $dir
		$data set label [file tail $dir]
		
		$m_tree insert end root [$data uuid] -text [$data get label] -image [Bitmap::get folder] -data $data
		
		wm withdraw .project
	}
	
	method editor { {e {}} } {
		if {$e != {}} {
			set m_editor $e
		}
		return $m_editor
	}
	
	method listProjects { } {
		$m_tree delete [$m_tree nodes root]
		
		foreach dir [glob -nocomplain "[config path.projects]/*"] {
			set data [Data new]
			$data set type folder
			$data set path $dir
			$data set label [file tail $dir]
		
			$m_tree insert end root [$data uuid] -text [$data get label] -image [Bitmap::get folder] -data $data
		}
	}
	
	method open { rx ry node } {
		$m_tree selection set $node
		set data [$m_tree itemcget $node -data]
		if {[$data get type] == "folder"} {
			set m_active $data
			foreach f [glob -nocomplain [$data get path]/*.dd] {
				set data [Data new]
				$data set path $f
				$data set label [file tail $f]
				$data set type file
				$m_tree insert end [$m_active uuid] [$data uuid] -text [$data get label] -image [Bitmap::get file] -data $data
			}
			$m_tree opentree $node
		} elseif {[$data get type] == "file"} {
			$m_editor load [$data get path]
		}
	}
	
	method popup { rx ry node } {
		$m_tree selection set $node
		set data [$m_tree itemcget $node -data]
		if {[$data get type] == "folder"} {
			set m_active $data
			tk_popup .menu_project $rx $ry
		}
	}
	
	method newFileDialog { } {
		set file [tk_getSaveFile -initialdir [$m_active get path]]
		if {$file == {}} {
			return
		}
		
		set fd [open $file w]
		close $fd
		set data [Data new]
		$data set path $file
		$data set label [file tail $file]
		$data set type file
		$m_tree insert end [$m_active uuid] [$data uuid] -text [$data get label] -image [Bitmap::get file] -data $data
		
		$m_editor load $file
	}
}

oo::class create Editor {
	variable m_app
	variable m_editor
	variable m_file
	variable m_outfile
	variable m_view
	
	constructor {app w} {
		set m_file {}
		set m_outfile {}
		set m_view [ImageViewer new]
	
		set m_app $app
		set m_editor $w
		
		bind . <Control-s> "[self] save"
		bind . <Control-r> "[self] run"
	}
	
	method load { file } {
		set m_outfile {}
		set m_file $file
		$m_app setFileName [file tail $file]
		
		set fd [open $m_file]
		set buf [read $fd]
		close $fd
		
		$m_editor delete 0.0 end
		$m_editor insert end $buf
	}
	
	method save { } {
		if {$m_file == {}} {
			return
		}
		
		$m_app clear
		
		$m_app verbose "saving...$m_file\n"
		set fd [open $m_file w]
		puts $fd [$m_editor get 0.0 end-1char]
		close $fd
	}
	
	method run { } {
		if {$m_file == {}} {
			return
		}
		
		$m_app clear
	
		$m_app verbose "loading ...\n"
		set cwd [pwd]
		set dir [file dirname $m_file]
		set file [file tail $m_file]
		set ext [string first . $file]
		if {$ext <= 0} {
			return
		}
		set name [string range $file 0 [expr $ext-1]]
		
		# read
		cd $dir
		set fd [open $file]
		set buf [read $fd]
		close $fd
		
		# process
		$m_app verbose "preprocessing ...\n"
		set dot [preprocess $buf]
		
		# create dot file
		$m_app verbose "creating dot file ...\n"
		set m_outfile ${name}.png
		set dotfile ${name}.dot
		set fd [open $dotfile w]
		fconfigure $fd -encoding utf-8
		puts $fd $dot
		close $fd
		
		# exec
		$m_app verbose "doing dot ...\n"
		set fd [open "|\"[config dot]\" -Tpng $dotfile -o $m_outfile"]
		while {[eof $fd] == 0} {
			gets $fd aline
			$m_app verbose $aline
		}
		close $fd
		
		# move directory
		cd $cwd
		
		$m_app verbose "done.\n"
		
		[self] reload
	}
	
	method reload { } {
		if {$m_outfile == {}} {
			return
		}
		set imgfile [file dirname $m_file]/$m_outfile
		
		$m_view show $imgfile
	}
	
	method showImageView { } {
		$m_view show
	}
}

oo::class create App {
	variable m_project
	variable m_editor
	variable m_console
	variable m_fileLabelFrame

	method create { } {
		pack [frame .mb] -fill x
		
		set m .mb.file.m
		pack [menubutton .mb.file -text "Project" -menu $m] -side left
		menu $m -tearoff 0
		$m add command -label "New" -command "[self] newProject"
		$m add separator
		$m add command -label "Quit" -command ""
		set m .mb.window.m
		pack [menubutton .mb.window -text "Window" -menu $m] -side left
		menu $m -tearoff 0
		$m add command -label "Image" -command "[self] showImageView"
		
		pack [frame .p1] -fill both -expand 1
		pack [labelframe .p1.w1 -text "Projects"] -side left -fill y
		Tree .p1.w1.t -xscrollcommand ".p1.w1.x set" -yscrollcommand ".p1.w1.y set" 
		scrollbar .p1.w1.x -command ".p1.w1.t xview" -orient horizontal
		scrollbar .p1.w1.y -command ".p1.w1.t yview"
		grid .p1.w1.t -row 0 -column 0 -sticky news
		grid .p1.w1.x -row 1 -column 0 -sticky ew
		grid .p1.w1.y -row 0 -column 1 -sticky ns
		grid columnconfigure .p1.w1 0 -weight 1
		grid rowconfigure .p1.w1 0 -weight 1
		set m_project [Project new [self] .p1.w1.t]
		pack [labelframe .p1.w2 -text "File:"] -side left -fill both -expand 1
		ctext .p1.w2.t -xscrollcommand ".p1.w2.x set" -yscrollcommand ".p1.w2.y set" -linemap 0 -height 30 -wrap none
		scrollbar .p1.w2.x -command ".p1.w2.t xview" -orient horizontal
		scrollbar .p1.w2.y -command ".p1.w2.t yview"
		grid .p1.w2.t -row 0 -column 0 -sticky news
		grid .p1.w2.x -row 1 -column 0 -sticky ew
		grid .p1.w2.y -row 0 -column 1 -sticky ns
		grid columnconfigure .p1.w2 0 -weight 1
		grid rowconfigure .p1.w2 0 -weight 1
		set m_editor [Editor new [self] .p1.w2.t]
		
		# setting tabs
		set font [.p1.w2.t cget -font]
		.p1.w2.t configure -tabs "[expr {4 * [font measure $font 0]}] left" -tabstyle wordprocessor
		
		set m_fileLabelFrame .p1.w2
		
		$m_project editor $m_editor
		
		pack [frame .p2] -fill x
		pack [text .p2.t -yscrollcommand ".p2.y set" -fg white -bg black -height 7] -side left -fill x -expand 1
		pack [scrollbar .p2.y -command ".p2.t yview"] -side left -fill y
		set m_console .p2.t
		$m_console tag configure PLAIN -foreground white -background black
		$m_console tag configure WARNING -foreground yellow
		$m_console tag configure ERROR -foreground orange
		$m_console tag configure INFO -foreground skyblue
		$m_console tag configure VERBOSE -foreground lightgreen
		
		$m_project listProjects
	}
	
	method newProject { } {
		$m_project showDialog
	}
	
	method setFileName { name } {
		$m_fileLabelFrame configure -text "File: $name"
	}
	
	method showImageView { } {
		$m_editor showImageView
	}
	
	method clear { } {
		$m_console delete 0.0 end
		update
	}
	
	method print { msg } {
		$m_console insert end $msg
		update
	}
	
	method warning { msg } {
		$m_console insert end $msg WARNING
		update
	}
	
	method error { msg } {
		$m_console insert end $msg ERROR
		update
	}
	
	method info { msg } {
		$m_console insert end $msg INFO
		update
	}
	
	method verbose { msg } {
		$m_console insert end $msg VERBOSE
		$m_console yview end
		update
	}
}

set app [App new]
$app create

source ko2dd_preprocessor.tcl

#set iv [ImageViewer new]
#$iv show ./projects/A/a2.png

